///////////////////////////////////////////////////////////
// 2D Voronoi Image Processing OpenGL/Glut 2D Demo.
// Written by Vladeta Stojanovic.
//
// Version: 020712_01
///////////////////////////////////////////////////////////

// Todo: Add Voronoi detail convergane algorithm as sampled ponts approach the border area of the target points

#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <glut.h>
#include <windows.h>
#include <iostream>
#include "auxMaths.h"
#include "voronoi.h"
#include "timer.h"

///////////////////////////////////////////////
//Global application variable declerations
///////////////////////////////////////////////
int gameWindow;
GLsizei iWidth = 1960; 
GLsizei iHeight = 1024; 
float scaleFactor = 1.0f;
float scrollX = 0;
float scrollY = 0;
int _time;
int _elapsedTime;
int delta;

////////////////////////////////////////////////////////
//Instanced global game object variable declerations
////////////////////////////////////////////////////////
Voronoi		*gVoronoi;
timer		*gTimer;

////////////////////////
//Function prototypes
////////////////////////
void DrawScene();
void DrawInfo();
void IdleScene();
void Visible(int vis);
void ResizeWindow(GLsizei w, GLsizei h);
void KeyboardInput(unsigned char cKey, int iX, int iY);
void DebugInfo();

void DebugInfo()
{
	//send any debug info to the application console via cout here
	//cout << delta * 0.005 << endl;
}

///////////////////////
//Main program loop
///////////////////////
int main(int argc, char **argv) 
{ 
	/////////////////////////////////
	//instatiate any objects here
	/////////////////////////////////
	gVoronoi = new Voronoi();

	//gVoronoi->LoadRAWMap("data/face_binary_edges.raw", 996);
	gVoronoi->LoadBMP("data/small_test.raw", 128);

	gTimer = new timer();

	//Disable this when analysing shapefile data
	iWidth = gVoronoi->m_Size;
	iHeight = gVoronoi->m_Size;

	////////////////////////////
	//Setup OpenGL/GLUT here
	////////////////////////////
    glutInit(&argc, argv);
    glutInitWindowPosition( 0, 0);
    glutInitWindowSize(iWidth, iHeight);
    glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE );
    glutCreateWindow( "2D Voronoi Topology Generation Demo");
	//glutReshapeFunc(ResizeWindow);
    glutDisplayFunc(DrawScene);
	glutVisibilityFunc(Visible);
    glutIdleFunc(IdleScene);
    glutKeyboardFunc(KeyboardInput);

	gVoronoi->InitProjection();

	//Default OpenGl window projection. When rendering shapefiles, override with custom shpefile bounding box based ortho bounds. 
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, iWidth, 0, iHeight); //adjust this if needed, directly effects the tiling of the generated texture
	glMatrixMode(GL_MODELVIEW); 
	glLoadIdentity();


	glClearColor (0.1, 0.57, 0.8, 1.0);
	gameWindow = glutGetWindow();

    glutMainLoop(); 

	return 0;
}

///////////////////////////////////////////////
//Call any critical updates in this function
///////////////////////////////////////////////
void IdleScene()
{
	_time = glutGet(GLUT_ELAPSED_TIME);
	delta  = _time - _elapsedTime;

	//critical updates go here
	{

	}

	_elapsedTime = _time;

	//Call debuig info here
	DebugInfo();

	glutPostWindowRedisplay(gameWindow);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//Resize/reshape window function (attempting to take a screenshot with this enabled messes things up, a lot. Figure out why).
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void ResizeWindow(GLsizei w, GLsizei h)
{
	iWidth = w;
    iHeight = h;

  /*  glViewport(0, 0, iWidth, iHeight);
  
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluOrtho2D(0, iWidth, 0, iHeight);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();*/

    //glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

//////////////////////////////
//Scene rendering function
//////////////////////////////
void DrawScene()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();

	glPushMatrix();
	{
		//Draw the tesselated Voronoi topology here
		glLoadIdentity();

		glScalef(scaleFactor, scaleFactor, 1.0);
		glTranslatef(scrollX, scrollY, 0);
		
		//use this for face images
		gVoronoi->DrawPoint(gVoronoi->m_Size / 2, true); //change this to 1 (defautl size is 16, but try with sized 4 and 8 as well).

		//use this for GIS images
		//gVoronoi->DrawPoint(16, true);
	}
	glPopMatrix();

	glutSwapBuffers(); 
}

void Visible(int vis)
{
	if(vis == GLUT_VISIBLE) 
	{
		glutIdleFunc(IdleScene);
	}
}

void KeyboardInput(unsigned char cKey, int iX, int iY)
{
	switch(cKey)
	{
	case '1': //Binary
		{
			gTimer->start();
			gVoronoi->GenerateAbs(255, false, 1);
			gTimer->stop();
			system("CLS");
			cout << "Done! " << endl;
			cout << "Total number of points computed: " << gVoronoi->m_TotalPoints << endl;
			cout << "Time taken (seconds): " << gTimer->elapsed_time() << endl;
			break;
		}
	case '2': //Range
		{
			gTimer->start();
			gVoronoi->GenerateRange(230, 255, 1);
			gTimer->stop();
			system("CLS");
			cout << "Done! " << endl;
			cout << "Total number of points computed: " << gVoronoi->m_TotalPoints << endl;
			cout << "Time taken (seconds): " << gTimer->elapsed_time() << endl;
			break;
		}

	case '3': //GIS data..implement sometime
		{
			break;
		}

	case '4' :// Save result
		{
			cout << "Result image outputed as result.bmp" << endl;
			gVoronoi->GenerateScreenshot("result.bmp", iWidth, iHeight);
			break;
		}

	case 'c': //Fill
		{
			cout << "Color fill display mode" << endl;
			glEnable(GL_CULL_FACE);
			glPolygonMode(GL_FRONT, GL_FILL);
			break;
		}

	case 'v': //wireframe
		{
			cout << "Wireframe display mode" << endl;

			//Probably a good idea to turn off backface culling here...
			glEnable(GL_CULL_FACE);
			glPolygonMode(GL_FRONT, GL_LINE);
			break;
		}

	//image transformation and scaling
	case 'w' :
		{
			scrollY += 10.0f;
			break;
		}
	case 's' :
		{
			scrollY -= 10.0f;
			break;
		}
	case 'a' :
		{
			scrollX -= 10.0f;
			break;
		}
	case 'd' :
		{
			scrollX += 10.0f;
			break;
		}

	case 't' :
		{
			scrollY += 0.5f;
			break;
		}
	case 'g' :
		{
			scrollY -= 0.5f;
			break;
		}
	case 'f' :
		{
			scrollX -= 0.5f;
			break;
		}
	case 'h' :
		{
			scrollX += 0.5f;
			break;
		}

	case 'q':
		{
			scaleFactor += 0.05f;
			break;
		}
	case 'e':
		{
			scaleFactor -= 0.05f;
			break;
		}

		//super zoom function s(for dealing with large shapefile zooming operations
	case 'z':
		{
			scaleFactor += 1.0f;
			break;
		}
	case 'x':
		{
			scaleFactor -= 1.0f;
			break;
		}

	default:
		{
			cout << "Invalid input" << endl;
			break;
		}
	}
}



