/////////////////////////////////////////////////////////////
// Voronoi generation and handling class implementation.
// Written by Vladeta Stojanovic.
//
// Version: 170612_01
////////////////////////////////////////////////////////////

#include "voronoi.h"
#include <iostream>

using namespace std;

static int sampledX = 0;
static int sampledY = 0;

Voronoi::Voronoi()
{
	mPoints = NULL;
	m_heightData.m_Data = NULL;
	m_heightData.targetDataX = NULL;
	m_heightData.targetDataY = NULL;
	m_TotalPoints = 0;

	Init();
}

Voronoi::~Voronoi()
{
	UnloadRAWMap();
}

//RAW map loading
bool Voronoi::LoadRAWMap(char * file_name, int _size)
{
	FILE* _file = NULL; //pointer to file for loading in

	//check if data was already loaded
	if(m_heightData.m_Data)
	{
		UnloadRAWMap();
	}

	//Open the data file
	_file = fopen(file_name, "rb");
	if(_file == NULL)
	{
		cout << "Could not open RAW height data file" << endl;
		return false;	
	}

	//allocate memory for RAW file data
	m_heightData.m_Data = new unsigned char[_size * _size * 2];
	m_heightData.targetDataX = new unsigned char[_size * _size * 2];
	m_heightData.targetDataY = new unsigned char[_size * _size * 2];

	//if not...
	if(m_heightData.m_Data == NULL)
	{
		cout << "Could allocate memory for RAW height data file" << endl;
		return false;		
	}

	cout << "RAW file loaded" << endl;
	cout << "Press 1 to extract binary color range data" << endl;
	cout << "Press 2 to extract color data in the given minimum and maximum range" << endl;

	//read in the data
	fread(m_heightData.m_Data, _size * _size, 1, _file);

	//close the file
	fclose(_file);
	

	//Set the size data
	m_heightData.m_Size_ = _size;
	m_Size = m_heightData.m_Size_;

	return true;
}

//RAW map unloading
bool Voronoi::UnloadRAWMap()
{
	//check if data has been already loaded
	if(m_heightData.m_Data)
	{
		//delete the data
		delete[] m_heightData.m_Data;
		delete[] m_heightData.targetDataX;
		delete[] m_heightData.targetDataY;

		//reset the map dimensions
		m_heightData.m_Size_ = 0;
	}

	return true;
}


void Voronoi::Init()
{
	node* point;

    srand(time(NULL));

    while (mPoints) 
	{
        point = mPoints;
        mPoints = mPoints->next;
        free(point);
    }
}

void Voronoi::InitProjection()
{
	glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glPointSize(4.0);

	//gShapetest->ResizeGL();
}

void Voronoi::DrawPoint(int _size, bool drawPoints)
{
	int width = _size, height = _size;
	
    node* point;

    point = mPoints;
    while (point)
	{
        glPushMatrix();
        glColor3ub(point->r, point->g, point->b);
        glTranslatef(point->x, point->y, 0.0);
        glutSolidCone(width  > height ? width : height, 1.0, 32, 1);

		if(drawPoints)
		{
            glDepthFunc(GL_ALWAYS);
            glColor3ub(255, 0, 0);
            glBegin(GL_POINTS);
            glVertex2i(0, 0);
            glEnd();
            glDepthFunc(GL_LEQUAL);
		}

        glPopMatrix();
        point = point->next;
    }
}


void Voronoi::GenerateAbsTarget(int absValue)
{
	for(int i = 0; i < m_Size; ++i)
	{
		for(int j = 0; j < m_Size; ++j)
		{
				if((int)m_heightData.m_Data[(j * m_Size) + i] == absValue) 
				{
					cout << "Generating target points..." << endl;

					m_heightData.targetDataX[(j * m_Size) + i] = m_Size - i;
					m_heightData.targetDataY[(j * m_Size) + i] = m_Size - j;
				}
		}
	}
	
	cout << "Done!" << endl;
}

//Calculates a new Voronoi topology map using a single colour value (binary colour extraction)
void Voronoi::GenerateAbs(int absValue, bool outside, int stepSize)
{
	node* point;

	if(stepSize <= 0)
	{
		stepSize = 1;
	}

	//First, smaple and setup the point valeus from the noise value map
	for(int i = 0; i < m_Size;)
	{
		i += stepSize;

		for(int j = 0; j < m_Size;)
		{
			j += stepSize;

			//Perhaps the entire image file should be loaded on the GPU memory, rather then system memory,
			//and the computed result should just be passed back to the device rendering context window buffer.
			//if(!outside)
			//{
				if(((int)m_heightData.m_Data[(j * m_Size) + i] == absValue))
				{
					//stepSize = 1;

					point = (node*)malloc(sizeof(node)); //allocate memory on the GPU memory rather then on system memory
					point->next = mPoints; //this operation may be very very costly....

					point->x = i;
					point->y = j; 

					cout << "point found at x:" << i << ", y: " << j << endl;

					//cout << "Generating target points. " << endl;

					point->r = rand()%255;
					point->g = rand()%255;
					point->b = rand()%255;
					mPoints = point; //this is very computationally expensive as well

					m_TotalPoints += 1;
				}
				
			//}
			//else if(outside)
			//{
			//	if((int)m_heightData.m_Data[(j * m_Size) + i] != absValue) 
			//	{
			//		cout << "point found at x:" << i << ", y: " << j << endl;

			//		point = (node*)malloc(sizeof(node)); //allocate memory on the GPU memory rather then on system memory
			//		point->next = mPoints; //this operation may be very very costly....

			//		point->x = i;
			//		point->y = j; 

			//		//Make it so that the closer the empty point is to the sampled path point, the smaller the stepping size gets
			//		//(every 96 pixels)

			//		//Test
			//		if((point->x > 96 && point->x < m_Size - 96) && (point->y > 96 && point->y < m_Size - 96))
			//		{
			//			stepSize = 24;
			//		}
			//		if((point->x > 192 && point->x < m_Size - 192) && (point->y > 192 && point->y < m_Size - 192))
			//		{
			//			stepSize = 16;
			//		}
			//		if((point->x > 288 && point->x < m_Size - 288) && (point->y > 288 && point->y < m_Size - 288))
			//		{
			//			stepSize = 8;
			//		}

			//		cout << "point->x: " << point->x << endl;
			//		cout << "point->y: " << point->y << endl;

			//		point->r = rand()%256;
			//		point->g = rand()%256;
			//		point->b = rand()%256;
			//		mPoints = point; //this is very computationally expensive as well

			//		m_TotalPoints += 1;
			//	}
			//}
		}
	 }

    glutPostRedisplay();
}

//Calculates a new Voronoi topology map using between a range of values (0-255)
void Voronoi::GenerateRange(int minValue, int maxValue, int stepSize)
{
	node* point;

	for(int i = 0; i < m_Size; i++)
	{
		for(int j = 0; j < m_Size; j++)
		{
			if((int)m_heightData.m_Data[(j * m_Size) + i] > minValue && (int)m_heightData.m_Data[(j * m_Size) + i] <= maxValue)
			{
				cout << "point found at x:" << i << ", y: " << j << endl;

				point = (node*)malloc(sizeof(node));
				point->next = mPoints;

				point->x = m_Size - i;
				point->y = m_Size - j; 

				cout << "point->x: " << point->x << endl;
				cout << "point->y: " << point->y << endl;

				point->r = rand()%256;
				point->g = rand()%256;
				point->b = rand()%256;
				mPoints = point;

				m_TotalPoints += 1;
			}
		}
	 }

    glutPostRedisplay();
}

//Implement 24bit BMP to RAW image conversion
void Voronoi::LoadBMP(char * file_name, int _size)
{
	FIBITMAP *dib = FreeImage_Load(FIF_BMP, file_name, 0);

	FIBITMAP *src = FreeImage_ConvertTo8Bits(dib);

	FreeImage_Unload(dib);

	BYTE *bits = (BYTE*)malloc(_size * _size);

	FreeImage_ConvertToRawBits(bits, src, _size, 8,
	FI_RGBA_RED, FI_RGBA_GREEN, FI_RGBA_BLUE,
	TRUE);

	FreeImage_Save(FIF_BMP, src, "temp/convert_temp.raw", 0);

	FreeImage_Unload(src);

	LoadRAWMap(file_name, _size);
}

//Output result function implementation
void Voronoi::GenerateScreenshot(char * file_name, int width, int height)
{
	BYTE* pixels = new BYTE[ 3 * width * height];

	glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, pixels);

	// Convert to FreeImage format & save to file
	FIBITMAP* image = FreeImage_ConvertFromRawBits(pixels, width, height, 3 * width, 24, 0x0000FF, 0xFF0000, 0x00FF00, true);
	FreeImage_Save(FIF_BMP, image, file_name, 0);

	// Free memory
	FreeImage_Unload(image);
	delete pixels;
}
